"""
Для решения FunCaptcha лучше всего использовать FirstCaptcha, OneCaptcha или CapSolver
"""
import asyncio
import time

from playwright.async_api import Error as PlaywrightError, TimeoutError as PlaywrightTimeoutError
from playwright.async_api import async_playwright, Browser, ProxySettings, Page, Playwright
from playwright_stealth import stealth_async
from yarl import URL

from .account import TwitterAccount

TWITTER_FUNCAPTCHA_SITEKEY = "0152B4EB-D2DC-460A-89A1-629838B529C9"
TWITTER_FUNCAPTCHA_URL = "https://twitter.com/account/access"


def proxy_url_to_playwright_proxy(proxy: str) -> ProxySettings:
    proxy = URL(proxy)
    return ProxySettings(
        server=f"{proxy.scheme}://{proxy.host}:{proxy.port}",
        password=proxy.password,
        username=proxy.user,
    )


class TwitterAccountUnlocker:
    def __init__(
            self,
            *,
            headless=False,
            default_proxy: str = None,
    ):
        self._playwright: Playwright | None = None
        self._browser: Browser | None = None
        self._headless = headless
        self._default_proxy = proxy_url_to_playwright_proxy(default_proxy) if default_proxy else None

    async def create_browser(self):
        self._playwright = await async_playwright().start()
        self._browser = await self._playwright.chromium.launch(
            headless=self._headless, proxy=self._default_proxy)

    async def __aenter__(self):
        await self.create_browser()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._browser.close()
        await self._playwright.stop()

    async def solve_captcha(self, page: Page, solved_captcha: str) -> bool:
        element = await page.query_selector("#arkose_iframe, input[type='submit'].Button.EdgeButton.EdgeButton--primary")

        if element and element.get_attribute("value") == "Continue to Twitter":
            await element.click()
            print(f'Account successfully unfrozen')

        if element and element.get_attribute('value') == 'Delete':
            await element.click()
            print(f'click delete')

        if element.get_attribute('value') == 'Start':
            await element.click()

            await page.goto('https://twitter.com/account/access')
            await page.wait_for_selector('#arkose_iframe')

        iframe_element = await page.query_selector('#arkose_iframe')
        if not iframe_element:
            if "twitter.com/home" in page.url:
                return True

        iframe = iframe_element.content_frame()
        script = f'parent.postMessage(JSON.stringify({{eventId:"challenge-complete",payload:{{sessionToken:"{solved_captcha}"}}}}),"*")'
        await iframe.evaluate(script)
        await page.wait_for_load_state(state='networkidle', timeout=10000)

    @staticmethod
    def wait_for_url(page, url, timeout=60):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if url in page.url:
                return True
            time.sleep(1)
        return False

    async def wait_for_multiple_conditions(self, page, selector, url, timeout=60000) -> tuple[any, any]:
        try:
            start_time = time.time()
            while time.time() - start_time < timeout / 1000:
                element = await page.query_selector(selector)
                if element:
                    return None, element

                if self.wait_for_url(page, url, 1):
                    return True, None

            return None, None

        except (PlaywrightError, PlaywrightTimeoutError) as e:
            return None, None

    async def unlock(
            self,
            account: TwitterAccount,
            *,
            proxy: str = None,
            default_timeout: int = 10,
    ):
        proxy = proxy_url_to_playwright_proxy(proxy) if proxy else None
        context = await self._browser.new_context(proxy=proxy)
        cookies = [
            {
                "name": "auth_token",
                "value": account.auth_token,
                "domain": "twitter.com",
                "path": "/",
            },
        ]
        await context.add_cookies(cookies)
        page = await context.new_page()
        page.set_default_timeout(default_timeout * 1000)
        await stealth_async(page)

        await page.goto(f"https://twitter.com/account/access")
        await page.wait_for_load_state(state='networkidle')

        if "access" not in page.url:
            return

        arkose_iframe = await page.query_selector("#arkose_iframe")
        button = await page.wait_for_selector('//*[@id="root"]/div/div[1]/button')
        await button.click()

        ...

        await context.close()



#structure of the Resp object (Resp is short for response)
'''
resp = Resp(response)
resp.raw        #returns raw (dict, decompressed) response from gateway
resp.event...   #handles event checking (resp.event.ready will be True if the response is of type READY)
resp.parsed...  #handles automatically parsing responses, has specific functions for specific types of responses (messages are handled differently from guild member list updates)
                #also has the ability to update bot objects like session
'''

from .event import Event
from .parse import Parse

class Resp:
	__slots__ = ['raw', 'event', 'parsed']
	def __init__(self, response):
		self.raw = response
		self.event = Event(response)
		self.parsed = Parse(response)


#combo
import time, datetime
class UserCombo(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def getCurrentUnixTs(self):
		date = datetime.datetime.now()
		unixts = int(time.mktime(date.timetuple())*1000) #milliseconds
		return unixts

	def constructEmojiDict(self, emoji, animatedEmoji):
		if emoji != None:
			if ":" in emoji:
				name, ID = emoji.split(":")
			else:
				name = emoji
				ID = None
			emojiDict = {}
			emojiDict.update({"id": ID, "name": name, "animated": animatedEmoji})
		else: #None
			emojiDict = None
		return emojiDict

	def constructActivitiesList(self, updates={}, remove=None): #update is a bit too raw, will fix later
		currentActivities = self.gatewayobj.session.userSettings["activities"]		
		activities = {4: {}, 0: {}, 1: {}, 2: {}, 3: {}}
		for i in currentActivities: #all this does is fix the format of current activities
			if remove!=i: #this helps remove activities
				activities[i] = dict(currentActivities[i])
				activities[i].pop("created_at", None)
				activities[i].pop("id", None)
				if "emoji" in activities[i]:
					if activities[i]["emoji"]!=None:
						emojiState = dict(currentActivities[i]["emoji"])
						name = emojiState.pop("name", "")
						ID = emojiState.pop("id", None)
						emojiStr = name+":"+ID if ID!=None else name
						animatedEmoji = emojiState.pop("animated", False)
					else:
						emojiStr = None
						animatedEmoji = False
					activities[i]["emoji"] = self.constructEmojiDict(emojiStr, animatedEmoji)
		for i in updates: #this adds activities
			activities[i] = updates[i]
		activitiesList = [activities[j] for j in activities if len(activities[j])>0]
		return activitiesList

	def setStatus(self, status):
		if status != self.gatewayobj.session.userSettings["status"]:
			self.gatewayobj.request.setStatus(status, self.constructActivitiesList())

	def setPlayingStatus(self, game): #will add metadata later
		currentStatus = self.gatewayobj.session.userSettings["status"]
		updates = {
		    0: {
		        "type": 0,
		        "name": game,
		        "timestamps": {"start": self.getCurrentUnixTs()}
		    }
		}
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates=updates))

	def removePlayingStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=0))

	def setStreamingStatus(self, stream, url): #will add metadata later
		currentStatus = self.gatewayobj.session.userSettings["status"]
		updates = {
		    1: {
		        "type": 1,
		        "name": stream,
		        "url": url,
		        "timestamps": {"start": self.getCurrentUnixTs()}
		    }
		}
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates=updates))

	def removeStreamingStatus(self): 
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=1))

	def setListeningStatus(self, song): #will add metadata later
		currentStatus = self.gatewayobj.session.userSettings["status"]
		updates = {
		    2: {
		        "type": 2,
		        "name": song,
		        "timestamps": {"start": self.getCurrentUnixTs()}
		    }
		}
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates=updates))

	def removeListeningStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=2))

	def setWatchingStatus(self, show): #will add metadata later
		currentStatus = self.gatewayobj.session.userSettings["status"]
		updates = {
		    3: {
		        "type": 3,
		        "name": show,
		        "timestamps": {"start": self.getCurrentUnixTs()}
		    }
		}
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates=updates))

	def removeWatchingStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=3))

	def setCustomStatus(self, customstatus, emoji, animatedEmoji):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		updates = {
		    4: {
		        "type": 4,
		        "state": customstatus,
		        "name": "Custom Status",
		        "emoji": self.constructEmojiDict(emoji, animatedEmoji),
		    }
		}
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates=updates))

	def removeCustomStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=4))

	def clearActivities(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, [])



from ..importmanager import Imports
imports = Imports(
	{
		"GuildRequest": "discum.gateway.guild.request",
		"DmRequest": "discum.gateway.dms.request",
		"UserRequest": "discum.gateway.user.request",
		"MediaRequest": "discum.gateway.media.request",
	}
)

class Request(object):
	__slots__ = ['gatewayobject']
	def __init__(self, gatewayobject):
		self.gatewayobject = gatewayobject #remember that the requests session obj is also passed in here

	def setStatus(self, status, activities=[], afk=False, since=0):
		imports.UserRequest(self.gatewayobject).setStatus(status, activities, afk, since)

	def lazyGuild(self, guild_id, channel_ranges=None, typing=None, threads=None, activities=None, members=None, thread_member_lists=None):
		imports.GuildRequest(self.gatewayobject).lazyGuild(guild_id, channel_ranges, typing, threads, activities, members, thread_member_lists)

	def searchGuildMembers(self, guild_ids, query="", limit=10, presences=True, user_ids=None, nonce=None):
		imports.GuildRequest(self.gatewayobject).searchGuildMembers(guild_ids, query, limit, presences, user_ids, nonce)

	def searchSlashCommands(self, guild_id, query=None, command_ids=None, applicationID=None, limit=10, offset=None, nonce="calculate", appType="chat"):
		imports.GuildRequest(self.gatewayobject).searchSlashCommands(guild_id, query, command_ids, applicationID, limit, offset, nonce, appType)

	def DMchannel(self, channel_id):
		imports.DmRequest(self.gatewayobject).DMchannel(channel_id)

	def call(self, channelID, guildID=None, mute=False, deaf=False, video=False):
		imports.MediaRequest(self.gatewayobject).call(channelID, guildID, mute, deaf, video)

	def endCall(self):
		imports.MediaRequest(self.gatewayobject).endCall()


from ..types import Types
#parses response from gateway

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class GuildParse(object):
	@staticmethod
	def guild_member_list_update(response):
		memberdata = {
		    "online_count": response["d"]["online_count"],
		    "member_count": response["d"]["member_count"],
		    "id": response["d"]["id"],
		    "guild_id": response["d"]["guild_id"],
		    "hoisted_roles": response["d"]["groups"],
		    "types": [],
		    "locations": [],
		    "updates": []
		}
		#now time to look over ops
		for chunk in response['d']['ops']:
			memberdata['types'].append(chunk['op'])
			if chunk['op'] in ('SYNC', 'INVALIDATE'):
				memberdata['locations'].append(chunk['range'])
				if chunk['op'] == 'SYNC':
					memberdata['updates'].append(chunk['items'])
				else: #invalidate
					memberdata['updates'].append([])
			elif chunk['op'] in ('INSERT', 'UPDATE', 'DELETE'): #only update the 0,99 range btw
				memberdata['locations'].append(chunk['index'])
				if chunk['op'] == 'DELETE':
					memberdata['updates'].append([])
				else:
					memberdata['updates'].append(chunk['item'])
		#and that's it. mostly some renaming and shuffling around to help with usage and reading :)
		return memberdata

	@staticmethod
	def guild_create(response, my_user_id):
		guilddata = dict(response["d"])
		#take care of position
		guilddata["my_data"] = response["d"].get("members", [])
		guilddata["members"] = {} #we dont actually get sent the member list from guild creates. however, this usually contains our position/role in that guild so...still good info
		guilddata["my_data"] = next((a for a in guilddata["my_data"] if a["user"]["id"]==my_user_id), {})
		if len(guilddata["my_data"]) == 1:
			guilddata["my_data"][0].pop("user", None)
			guilddata["my_data"][0]["user_id"] = my_user_id
		#take care of emojis
		guilddata["emojis"] = {i["id"]:i for i in response["d"]["emojis"]}
		#take care of roles
		guilddata["roles"] = {j["id"]:j for j in response["d"]["roles"]}
		#take care of channels
		guilddata["channels"] = {k["id"]:dict(k, **{"type":Types.CHANNEL[k["type"]]}) for k in response["d"]["channels"]}
		return guilddata

	@staticmethod
	def guild_members_chunk(response):
		memberChunkData = {"members":[], "guild_id": response["d"]["guild_id"], "chunk_count":response["d"]["chunk_count"], "chunk_index":response["d"]["chunk_index"]}
		if "not_found" in response["d"]:
			memberChunkData["not_found"] = [str(n) for n in response["d"]["not_found"]] #list of user ids
		presences = {}
		if "presences" in response["d"]:
			presences = {i["user"]["id"]:i for i in response["d"]["presences"]}
		for user in response["d"]["members"]:
			completeData = dict(user) 
			defaultPresence = {"user": {"id": user.get("user").get("id")}, "status": "offline", "client_status": {}, "activities": []} #offline status
			completeData["presence"] = presences.pop(user.get("user").get("id"), defaultPresence)
			memberChunkData["members"].append(completeData)
		return memberChunkData



class Event:
	__slots__ = ['response']
	def __init__(self, response):
		self.response = response

	@property
	def ready(self):
		return self.response['t'] == 'READY'

	@property
	def ready_supplemental(self):
		return self.response['t'] == 'READY_SUPPLEMENTAL'

	@property
	def activity_join_request(self): #not sure what this is
		return self.response['t'] == 'ACTIVITY_JOIN_REQUEST'

	@property
	def activity(self): #not sure what this is either
		return self.response['t'] == 'ACTIVITY_START'

	@property
	def braintree(self):
		return self.response['t'] == 'BRAINTREE_POPUP_BRIDGE_CALLBACK'

	@property
	def call(self):
		return self.response['t'] == 'CALL_CREATE'

	@property
	def call_deleted(self):
		return self.response['t'] == 'CALL_DELETE'

	@property
	def call_updated(self):
		return self.response['t'] == 'CALL_UPDATE'

	@property
	def channel(self):
		return self.response['t'] == 'CHANNEL_CREATE'

	@property
	def channel_deleted(self):
		return self.response['t'] == 'CHANNEL_DELETE'

	@property
	def channel_updated(self):
		return self.response['t'] == 'CHANNEL_UPDATE'

	@property
	def channel_read_state_updated(self): #{'t': 'CHANNEL_UNREAD_UPDATE', 's': s, 'op': 0, 'd': {'guild_id': '', 'channel_unread_updates': [{'last_message_id': '', 'id': 'unread id i guess'}, ...]}}
		return self.response['t'] == 'CHANNEL_UNREAD_UPDATE'

	@property
	def pins_ack(self):
		return self.response['t'] == 'CHANNEL_PINS_ACK'

	@property
	def pins_updated(self):
		return self.response['t'] == 'CHANNEL_PINS_UPDATE'

	@property
	def recipient_added(self):
		return self.response['t'] == 'CHANNEL_RECIPIENT_ADD'

	@property
	def recipient_removed(self):
		return self.response['t'] == 'CHANNEL_RECIPIENT_REMOVE'

	@property
	def entitlement(self):
		return self.response['t'] == 'ENTITLEMENT_CREATE'

	@property
	def entitlement_deleted(self):
		return self.response['t'] == 'ENTITLEMENT_DELETE'

	@property
	def entitlement_updated(self):
		return self.response['t'] == 'ENTITLEMENT_UPDATE'

	@property
	def friend_suggestion(self):
		return self.response['t'] == 'FRIEND_SUGGESTION_CREATE'

	@property
	def friend_suggestion_deleted(self):
		return self.response['t'] == 'FRIEND_SUGGESTION_DELETE'

	@property
	def gift_code_updated(self):
		return self.response['t'] == 'GIFT_CODE_UPDATE'

	@property
	def guild_application_commands_updated(self):
		return self.response['t'] == 'GUILD_APPLICATION_COMMANDS_UPDATE'

	@property
	def guild_application_command_counts_updated(self):
		return self.response['t'] == 'GUILD_APPLICATION_COMMAND_COUNTS_UPDATE'

	@property
	def ban_added(self): #{'t': 'GUILD_BAN_ADD', 's': s, 'op': 0, 'd': {'user': {'username': username, 'public_flags': 0, 'id': id, 'discriminator': '0000', 'avatar': None}, 'guild_id': guildID}}
		return self.response['t'] == 'GUILD_BAN_ADD'

	@property
	def ban_removed(self):
		return self.response['t'] == 'GUILD_BAN_REMOVE'

	@property
	def guild(self):
		return self.response['t'] == 'GUILD_CREATE'

	@property
	def guild_deleted(self):
		return self.response['t'] == 'GUILD_DELETE'

	@property
	def guild_updated(self):
		return self.response['t'] == 'GUILD_UPDATE'

	@property
	def emojis_updated(self):
		return self.response['t'] == 'GUILD_EMOJIS_UPDATE'

	@property
	def guild_integrations_updated(self):
		return self.response['t'] == 'GUILD_INTEGRATIONS_UPDATE'

	@property
	def integration(self):
		return self.response['t'] == 'INTEGRATION_CREATE'

	@property
	def integration_updated(self):
		return self.response['t'] == 'INTEGRATION_UPDATE'

	@property
	def integration_deleted(self):
		return self.response['t'] == 'INTEGRATION_DELETE'

	@property
	def interaction_successful(self):
		return self.response['t'] == 'INTERACTION_SUCCESS'

	@property
	def guild_member_list(self):
		return self.response['t'] == 'GUILD_MEMBER_LIST_UPDATE'

	@property
	def guild_member_updated(self): #{'t': 'GUILD_MEMBER_UPDATE', 's': s, 'op': 0, 'd': {'user': {'username': usernamd, 'public_flags': 0, 'id': id, 'discriminator': '0000', 'avatar': avatar}, 'roles': [], 'premium_since': None, 'pending': False, 'nick': None, 'joined_at': '', 'is_pending': False, 'guild_id': guildID}}
		return self.response['t'] == 'GUILD_MEMBER_UPDATE'

	@property
	def guild_members_chunk(self): #used for op 8 (when searching in the search bar and members come up)
		return self.response['t'] == 'GUILD_MEMBERS_CHUNK'

	@property
	def guild_member_removed(self): #only seems to get sent when you leave a guild (voluntarily or involuntarily)...but the server owner sees no such event? weird.
		return self.response['t'] == 'GUILD_MEMBER_REMOVE'

	@property
	def role(self):
		return self.response['t'] == 'GUILD_ROLE_CREATE'

	@property
	def role_deleted(self):
		return self.response['t'] == 'GUILD_ROLE_DELETE'

	@property
	def role_updated(self):
		return self.response['t'] == 'GUILD_ROLE_UPDATE'

	@property
	def invite(self):
		return self.response['t'] == 'INVITE_CREATE'

	@property
	def invite_deleted(self):
		return self.response['t'] == 'INVITE_DELETE'

	@property
	def library_app_updated(self):
		return self.response['t'] == 'LIBRARY_APPLICATION_UPDATE'

	@property
	def lobby(self): #idk...https://discord.com/developers/docs/game-sdk/lobbies...
		return self.response['t'] == 'LOBBY_CREATE'

	@property
	def lobby_deleted(self):
		return self.response['t'] == 'LOBBY_DELETE'

	@property
	def lobby_updated(self):
		return self.response['t'] == 'LOBBY_UPDATE'

	@property
	def lobby_member_connected(self):
		return self.response['t'] == 'LOBBY_MEMBER_CONNECT'

	@property
	def lobby_member_disconnected(self):
		return self.response['t'] == 'LOBBY_MEMBER_DISCONNECT'

	@property
	def lobby_member_updated(self):
		return self.response['t'] == 'LOBBY_MEMBER_UPDATE'

	@property
	def lobby_message(self):
		return self.response['t'] == 'LOBBY_MESSAGE'

	@property
	def lobby_voice_server_update(self):
		return self.response['t'] == 'LOBBY_VOICE_SERVER_UPDATE'

	@property
	def lobby_voice_state_update(self):
		return self.response['t'] == 'LOBBY_VOICE_STATE_UPDATE'

	@property
	def message_ack(self): #{'t': 'MESSAGE_ACK', 's': s, 'op': 0, 'd': {'version': v, 'message_id': '', 'channel_id': ''}}
		return self.response['t'] == 'MESSAGE_ACK'

	@property
	def message(self):
		return self.response['t'] == 'MESSAGE_CREATE'

	@property
	def message_deleted(self):
		return self.response['t'] == 'MESSAGE_DELETE'

	@property
	def bulk_messages_deleted(self): #{'t': 'MESSAGE_DELETE_BULK', 's': s, 'op': 0, 'd': {'ids': [], 'channel_id': '', 'guild_id': ''}}
		return self.response['t'] == 'MESSAGE_DELETE_BULK'

	@property
	def reaction_added(self):
		return self.response['t'] == 'MESSAGE_REACTION_ADD'

	@property
	def reaction_removed(self):
		return self.response['t'] == 'MESSAGE_REACTION_REMOVE'

	@property
	def all_message_reactions_removed(self):
		return self.response['t'] == 'MESSAGE_REACTION_REMOVE_ALL'

	@property
	def message_reaction_emoji_removed(self): #not entirely sure what's the difference between this and MESSAGE_REACTION_REMOVE but ok
		return self.response['t'] == 'MESSAGE_REACTION_REMOVE_EMOJI'

	@property
	def message_updated(self):
		return self.response['t'] == 'MESSAGE_UPDATE'

	@property
	def oauth2_token_removed(self): #maybe has to do with deleting an account?
		return self.response['t'] == 'OAUTH2_TOKEN_REMOVE'

	@property
	def presence_replaced(self):
		return self.response['t'] == 'PRESENCES_REPLACE'

	@property
	def presence_updated(self):
		return self.response['t'] == 'PRESENCE_UPDATE'

	@property
	def recent_mention_deleted(self):
		return self.response['t'] == 'RECENT_MENTION_DELETE'

	@property
	def relationship_added(self):
		return self.response['t'] == 'RELATIONSHIP_ADD'

	@property
	def relationship_removed(self):
		return self.response['t'] == 'RELATIONSHIP_REMOVE'

	@property
	def session_replaced(self):
		return self.response['t'] == 'SESSIONS_REPLACE'

	@property
	def stream(self):
		return self.response['t'] == 'STREAM_CREATE'

	@property
	def stream_deleted(self):
		return self.response['t'] == 'STREAM_DELETE'

	@property
	def stream_server_updated(self):
		return self.response['t'] == 'STREAM_SERVER_UPDATE'

	@property
	def stream_updated(self):
		return self.response['t'] == 'STREAM_UPDATE'

	@property
	def typing(self):
		return self.response['t'] == 'TYPING_START'

	@property
	def achievement_updated(self): #idk
		return self.response['t'] == 'USER_ACHIEVEMENT_UPDATE'

	@property
	def connections_updated(self):
		return self.response['t'] == 'USER_CONNECTIONS_UPDATE'

	@property
	def feed_settings_updated(self):
		return self.response['t'] == 'USER_FEED_SETTINGS_UPDATE'

	@property
	def user_guild_settings_updated(self): #like notifications for guilds you're in and that sort of stuff
		return self.response['t'] == 'USER_GUILD_SETTINGS_UPDATE'

	@property
	def note_updated(self):
		return self.response['t'] == 'USER_NOTE_UPDATE'

	@property
	def payment_sources_updated(self):
		return self.response['t'] == 'USER_PAYMENT_SOURCES_UPDATE'

	@property
	def payments_updated(self):
		return self.response['t'] == 'USER_PAYMENTS_UPDATE'

	@property
	def user_premium_guild_sub_slot(self): #now thats a long name
		return self.response['t'] == 'USER_PREMIUM_GUILD_SUBSCRIPTION_SLOT_CREATE'

	@property
	def user_premium_guild_sub_slot_updated(self):
		return self.response['t'] == 'USER_PREMIUM_GUILD_SUBSCRIPTION_SLOT_UPDATE'

	@property
	def required_action_updated(self): #idk
		return self.response['t'] == 'USER_REQUIRED_ACTION_UPDATE'

	@property
	def settings_updated(self): #user settings
		return self.response['t'] == 'USER_SETTINGS_UPDATE'

	@property
	def subscriptions_updated(self):
		return self.response['t'] == 'USER_SUBSCRIPTIONS_UPDATE'

	@property
	def stickers_updated(self):
		return self.response['t'] == 'USER_STICKER_PACK_UPDATE'

	@property
	def user_updated(self):
		return self.response['t'] == 'USER_UPDATE'

	@property
	def voice_server_updated(self):
		return self.response['t'] == 'VOICE_SERVER_UPDATE'

	@property
	def voice_state_updated(self):
		return self.response['t'] == 'VOICE_STATE_UPDATE'

	@property
	def webhooks_updated(self):
		return self.response['t'] == 'WEBHOOKS_UPDATE'

	@property
	def thread(self):
		return self.response['t'] == 'THREAD_CREATE'

	@property
	def thread_updated(self):
		return self.response['t'] == 'THREAD_UPDATE'

	@property
	def thread_deleted(self):
		return self.response['t'] == 'THREAD_DELETE'

	@property
	def thread_list_sync(self): #Sent when the current user gains access to a thread
		return self.response['t'] == 'THREAD_LIST_SYNC'

	@property
	def thread_member_updated(self): #Sent when the thread member object for the current user is updated
		return self.response['t'] == 'THREAD_MEMBER_UPDATE'

	@property
	def thread_members_updated(self): #Sent when anyone is added to or removed from a thread
		return self.response['t'] == 'THREAD_MEMBERS_UPDATE'



import websocket
import json
import time
import random
import zlib
import copy
from enum import IntEnum

try:
    import thread
except ImportError:
    import _thread as thread

from .session import Session
from .response import Resp
from .request import Request
from .parse import Parse

from discum.user.user import User
from .guild.combo import GuildCombo
from .user.combo import UserCombo

#exceptions
class InvalidSessionException(Exception):
    pass

class NeedToReconnectException(Exception):
    pass

class ConnectionResumableException(Exception): #for certain close codes. "exception"
    pass

class ConnectionManuallyClosedException(Exception):
    pass

def exceptionChecker(e, types): #this is an A or B or ... check
    for i in types:
        if isinstance(e,i):
            return True
    return False


class OPCODE(IntEnum):
    #                                    Client Action  Description
    DISPATCH =                      0  # Receive        dispatches an event
    HEARTBEAT =                     1  # Send/Receive   used for ping checking
    IDENTIFY =                      2  # Send           used for client handshake
    PRESENCE_UPDATE =               3  # Send           used to update the client status
    VOICE_STATE_UPDATE =            4  # Send           used to join/move/leave voice channels
    VOICE_SERVER_PING =             5  # Send           used for voice ping checking
    RESUME =                        6  # Send           used to resume a closed connection
    RECONNECT =                     7  # Receive        used to tell when to reconnect (sometimes...)
    REQUEST_GUILD_MEMBERS =         8  # Send           used to request guild members (when searching for members in the search bar of a guild)
    INVALID_SESSION =               9  # Receive        used to notify client they have an invalid session id
    HELLO =                        10  # Receive        sent immediately after connecting, contains heartbeat and server debug information
    HEARTBEAT_ACK =                11  # Sent           immediately following a client heartbeat that was received
    GUILD_SYNC =                   12  # Receive        guild_sync but not used anymore
    DM_UPDATE =                    13  # Send           used to get dm features
    LAZY_REQUEST =                 14  # Send           discord responds back with GUILD_MEMBER_LIST_UPDATE type SYNC...
    LOBBY_CONNECT =                15  # ??
    LOBBY_DISCONNECT =             16  # ??
    LOBBY_VOICE_STATES_UPDATE =    17  # Receive
    STREAM_CREATE =                18  # ??
    STREAM_DELETE =                19  # ??
    STREAM_WATCH =                 20  # ??
    STREAM_PING =                  21  # Send
    STREAM_SET_PAUSED =            22  # ??
    REQUEST_APPLICATION_COMMANDS = 24  # Send           request application/bot cmds (user, message, and slash cmds)


class GatewayServer:
    def __init__(
            self,
            api_version: int,
            auth_token: str,
            user_agent: str,
            x_super_properties, sessionobj=None, RESTurl=""): #session obj needed for proxies and some combo gateway functions (that also require http api wraps)
        self.websocket_url = f"wss://gateway.discord.gg/?encoding=json&v={api_version}&compress=zlib-stream"
        self.auth_token = auth_token
        self.x_super_properties = x_super_properties
        self.auth = {
            "token": self.auth_token,
            "capabilities": 509,
            "properties": self.x_super_properties,
            "presence": {
                "status": "online",
                "since": 0,
                "activities": [],
                "afk": False
            },
            "compress": False,
            "client_state": {
                "guild_hashes": {},
                "highest_last_message_id": "0",
                "read_state_version": 0,
                "user_guild_settings_version": -1,
                "user_settings_version": -1
            }
        }
        self.RESTurl = RESTurl #for helper http requests
        self.sessionobj = sessionobj #for helper http requests

        self.proxy_type, self.proxy_auth, self.proxy_host, self.proxy_port = [None]*4
        if sessionobj and sessionobj.proxies:
            self.proxy_type = proxy_type = list(sessionobj.proxies.keys())[0]
            self.proxy_host, self.proxy_port = sessionobj.proxies[proxy_type].split('://')[-1].split(':')
            if sessionobj.auth:
                self.proxy_auth = (sessionobj.auth.username, sessionobj.auth.password)

        self.keepData = ("guilds") #keep data even after leaving "dms", "guilds", or "guild_channels"

        self.interval = None
        self.session_id = None
        self.sequence = 0
        self.READY = False #becomes True once READY_SUPPLEMENTAL is received
        self.session = Session({},{}) #not the same as sessionobj

        #websocket.enableTrace(True) #for debugging
        self.ws = self._get_ws_app(self.websocket_url)

        self._after_message_hooks = []
        self._last_err = None
        self._last_close_event = None

        self.connected = False
        self.resumable = False

        self.voice_data = {} #voice connections dependent on current (connected) session

        self.memberFetchingStatus = {"first": []}
        self.resetMembersOnSessionReconnect = True #reset members after each session
        self.updateSessionData = True
        self.guildMemberSearches = {}

        #latency
        self._last_ack = None
        self.latency = None

        #gateway requests and parsing
        self.request = Request(self)
        self.parse = Parse

        #extra gateway connection kwargs
        self.connectionKwargs = {}

    # discum: https://github.com/websocket-client/websocket-client/blob/master/websocket/_app.py#L84
    def _get_ws_app(self, websocketurl):
        headers = {
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
            "Cache-Control": "no-cache",
            "Pragma": "no-cache",
            "Sec-WebSocket-Extensions": "permessage-deflate; client_max_window_bits",
            "User-Agent": self.x_super_properties["browser_user_agent"]
        } #more info: https://stackoverflow.com/a/40675547

        ws = websocket.WebSocketApp(websocketurl,
                                    header = headers,
                                    on_open=lambda ws: self.on_open(ws),
                                    on_message=lambda ws, msg: self.on_message(ws, msg),
                                    on_error=lambda ws, msg: self.on_error(ws, msg),
                                    on_close=lambda ws, close_code, close_msg: self.on_close(ws, close_code, close_msg)
                                    )
        return ws

    def decompress(self, bmessage): #input is byte message
        data = self._zlib.decompress(bmessage)
        jsonmessage = json.loads(data.decode("UTF8"))
        return jsonmessage

    def on_open(self, ws):
        self.connected = True
        self.memberFetchingStatus = {"first": []}
        Logger.log("[gateway] Connected to websocket.", None, self.log)
        if not self.resumable:
            #send presences if 1 or more activites in previous session. Whether or not you're invisible doesn't matter apparently.
            if len(self.session.settings_ready) != 0:
                if self.session.userSettings.get("activities") not in (None, {}):
                    self.auth["presence"]["status"] = self.session.userSettings.get("status")
                    self.auth["presence"]["activities"] = imports.UserCombo(self).constructActivitiesList()
            self.send({"op": self.OPCODE.IDENTIFY, "d": self.auth})
            self.send({"op": self.OPCODE.VOICE_STATE_UPDATE, "d": {"guild_id": None, "channel_id": None, "self_mute": True, "self_deaf": False, "self_video": False}})

        else:
            self.resumable = False
            self.send({"op": self.OPCODE.RESUME, "d": {"token": self.token, "session_id": self.session_id, "seq": self.sequence-1 if self.sequence>0 else self.sequence}})

    def on_message(self, ws, message):
        response = self.decompress(message)
        if response['op'] != self.OPCODE.HEARTBEAT_ACK:
            self.sequence += 1
        resp = Resp(copy.deepcopy(response))
        Logger.log('[gateway] < {}'.format(response), LogLevel.RECEIVE, self.log)
        if response['op'] == self.OPCODE.HELLO: #only happens once, first message sent to client
            self.interval = (response["d"]["heartbeat_interval"])/1000 #if this fails make an issue and I'll revert it back to the old method (slightly smaller wait time than heartbeat)
            thread.start_new_thread(self._heartbeat, ())
        elif response['op'] == self.OPCODE.HEARTBEAT_ACK:
            if self._last_ack != None:
                self.latency = time.perf_counter() - self._last_ack
        elif response['op'] == self.OPCODE.HEARTBEAT:
            self.send({"op": self.OPCODE.HEARTBEAT,"d": self.sequence})
        elif response['op'] == self.OPCODE.INVALID_SESSION:
            Logger.log("[gateway] Invalid session.", None, self.log)
            self._last_err = InvalidSessionException("Invalid Session Error.")
            if self.resumable:
                self.resumable = False
                self.sequence = 0
                self.close()
            else:
                self.sequence = 0
                self.close()
        elif response['op'] == self.OPCODE.RECONNECT:
            Logger.log("[gateway] Received opcode 7 (reconnect).", None, self.log)
            self._last_err = NeedToReconnectException("Discord sent an opcode 7 (reconnect).")
            self.close()
        if self.interval == None:
            Logger.log("[gateway] Identify failed.", None, self.log)
            self.close()
        if resp.event.ready:
            self._last_err = None
            self.session_id = response['d']['session_id']
            settings_ready = resp.parsed.ready() #parsed
            if not self.resetMembersOnSessionReconnect and self.session.read()[0]:
                for guildID in settings_ready['guilds']:
                    settings_ready['guilds'][guildID]['members'] = self.session.guild(guildID).members
            self.session.setSettingsReady(settings_ready)
        elif resp.event.ready_supplemental:
            settings_ready_supp = resp.parsed.ready_supplemental() #parsed
            self.session.setSettingsReadySupp(settings_ready_supp)
            self.READY = True
        if self.updateSessionData:
            self.sessionUpdates(resp)
        thread.start_new_thread(self._response_loop, (resp,))

    def on_error(self, ws, error):
        Logger.log('[gateway] < {}'.format(error), LogLevel.WARNING, self.log)
        self._last_err = error

    def on_close(self, ws, close_code, close_msg):
        self.connected = False
        self.READY = False #reset self.READY
        if close_code or close_msg:
            self._last_close_event = {"code": close_code, "reason": close_msg}
            Logger.log("[gateway] close status code: " + str(close_code), None, self.log)
            Logger.log("[gateway] close message: " + str(close_msg), None, self.log)
            if not (4000<close_code<=4010):
                self.resumable = True
                self._last_err = ConnectionResumableException("Connection is resumable.")
            if close_code in (None, 1000, 1001, 1006):
                self._last_err = ConnectionManuallyClosedException("Disconnection initiated by client using close function.")
        Logger.log('[gateway] websocket closed', None, self.log)

    #Discord needs heartbeats, or else connection will sever
    def _heartbeat(self):
        Logger.log("[gateway] entering heartbeat", None, self.log)
        while self.connected:
            if self.interval == None: #can't replicate the issue so consider this a temp patch
                self.interval = 41.25
            time.sleep(self.interval)
            if not self.connected:
                break
            self.send({"op": self.OPCODE.HEARTBEAT,"d": self.sequence})
            self._last_ack = time.perf_counter()
        return

    #just a wrapper for ws.send
    def send(self, payload):
        Logger.log('[gateway] > {}'.format(payload), LogLevel.SEND, self.log)
        self.ws.send(json.dumps(payload))

    def close(self):
        self.connected = False
        self.READY = False #reset self.READY
        if not exceptionChecker(self._last_err, [InvalidSessionException, NeedToReconnectException]):
            self._last_err = ConnectionManuallyClosedException("Disconnection initiated by client using close function.")
        Logger.log('[gateway] websocket closed', None, self.log) #don't worry if this message prints twice
        self.ws.close()

    def command(self, func):
        if callable(func):
            self._after_message_hooks.append(func)
            return func
        elif isinstance(func, dict): #because I can't figure out out to neatly pass params to decorators :(. Normal behavior still works; use as usual.
            priority = func.pop('priority', len(self._after_message_hooks))
            self._after_message_hooks.insert(priority, func)
            return func['function']

    #kinda influenced by https://github.com/scrubjay55/Reddit_ChatBot_Python/blob/master/Reddit_ChatBot_Python/WebSockClient.py (Apache License 2.0)
    def _response_loop(self, resp): #thx ToasterUwU for bringing up dummy threads
        commandslist = self._after_message_hooks[:] #create a copy
        for func in commandslist:
            if callable(func):
                func(resp)
            elif isinstance(func, dict):
                function = func['function']
                params = func['params'] if 'params' in func else {}
                function(resp, **params)
        return

    def removeCommand(self, func, exactMatch=True, allMatches=False):
        try:
            if exactMatch:
                self._after_message_hooks.index(func) #for raising the value error
                if allMatches:
                    self._after_message_hooks = [i for i in self._after_message_hooks if i!=func]
                else: #simply remove first found
                    del self._after_message_hooks[self._after_message_hooks.index(func)]
            else:
                commandsCopy = [i if callable(i) else i['function'] for i in self._after_message_hooks] #list of just functions
                commandsCopy.index(func) #for raising the value error
                if allMatches:
                    self._after_message_hooks = [i for (i,j) in zip(self._after_message_hooks, commandsCopy) if j!=func]
                else:
                    del self._after_message_hooks[commandsCopy.index(func)]
        except ValueError:
            Logger.log('{} not found in _after_message_hooks.'.format(func), None, self.log)
            pass

    def clearCommands(self):
        self._after_message_hooks = []

    def resetSession(self): #just resets some variables that in-turn, resets the session (client side). Do not run this while running run().
        self.interval = None
        self.session_id = None
        self.sequence = 0
        self.READY = False #becomes True once READY_SUPPLEMENTAL is received
        self._last_err = None
        self.voice_data = {}
        self.resumable = False #you can't resume anyways without session_id and sequence
        self._last_ack = None
        self.memberFetchingStatus = {"first": []}

    #kinda influenced by https://github.com/scrubjay55/Reddit_ChatBot_Python (Apache License 2.0)
    def run(self, auto_reconnect=True):
        if auto_reconnect:
            while True:
                try:
                    self._zlib = zlib.decompressobj()
                    self.ws.run_forever(
                        ping_interval=10,
                        ping_timeout=5,
                        http_proxy_host=self.proxy_host,
                        http_proxy_port=self.proxy_port,
                        http_proxy_auth=self.proxy_auth,
                        proxy_type=self.proxy_type,
                        **self.connectionKwargs
                    )
                    raise self._last_err
                except KeyboardInterrupt:
                    self._last_err = KeyboardInterrupt("Keyboard Interrupt Error")
                    Logger.log("[gateway] Connection forcibly closed using Keyboard Interrupt.", None, self.log)
                    break
                except Exception as e:
                    if auto_reconnect:
                        if not exceptionChecker(e, [KeyboardInterrupt]):
                            if exceptionChecker(e, [ConnectionResumableException]):
                                self._last_err = None
                                waitTime = random.randrange(1,6)
                                Logger.log("[gateway] Connection Dropped. Attempting to resume last valid session in {} seconds.".format(waitTime), None, self.log)
                                time.sleep(waitTime)
                            elif exceptionChecker(e, [ConnectionManuallyClosedException]):
                                Logger.log("[gateway] Connection forcibly closed using close function.", None, self.log)
                                break
                            else:
                                self.resetSession()
                                Logger.log("[gateway] Connection Dropped. Retrying in 10 seconds.", None, self.log)
                                time.sleep(10)
        else:
            self._zlib = zlib.decompressobj()
            self.ws.run_forever(
                ping_interval=10,
                ping_timeout=5,
                http_proxy_host=self.proxy_host,
                http_proxy_port=self.proxy_port,
                http_proxy_auth=self.proxy_auth,
                proxy_type=self.proxy_type,
                **self.connectionKwargs
            )

    ######################################################
    def sessionUpdates(self, resp):
        #***guilds
        #guild created
        if resp.event.guild:
            guildData = resp.parsed.guild_create(my_user_id=self.session.user['id']) #user id needed for updating personal roles in that guild
            guildID = guildData['id']
            voiceStateData = guildData.pop('voice_states', [])
            if not self.resetMembersOnSessionReconnect and guildID in self.session.guildIDs:
                guildData['members'] = self.session.guild(guildID).members
            self.session.setGuildData(guildID, guildData)
            self.session.setVoiceStateData(guildID, voiceStateData)
        #guild deleted
        elif resp.event.guild_deleted:
            if "guilds" in self.keepData:
                self.session.guild(resp.raw['d']['id']).updateData({"removed": True})  #add the indicator
            else:
                self.session.removeGuildData(resp.raw['d']['id'])

        #***channels (dms and guilds)
        #channel created (either dm or guild channel)
        elif resp.event.channel:
            channelData = resp.parsed.channel_create()
            channelID = channelData['id']
            if channelData["type"] in ("dm", "group_dm"): #dm
                self.session.setDmData(channelID, channelData)
            else: #other channels
                guildID = channelData.pop("guild_id")
                self.session.guild(guildID).setChannelData(channelID, channelData)
        #channel deleted (either dm or guild channel)
        elif resp.event.channel_deleted:
            channelData = resp.parsed.channel_delete() #updated data :) ...unlike guild_delete events
            channelData["removed"] = True #add the indicator
            channelID = channelData["id"]
            if channelData["type"] in ("dm", "group_dm"): #dm
                if "dms" in self.keepData:
                    self.session.DM(channelID).updateData(channelData)
                else:
                    self.session.removeDmData(channelID)
            else: #other channels (guild channels)
                guildID = channelData.pop("guild_id")
                if "guild_channels" in self.keepData:
                    self.session.guild(guildID).updateChannelData(channelID, channelData)
                else:
                    self.session.guild(guildID).removeChannelData(channelID)

        #***user updates
        #user settings updated
        elif resp.event.settings_updated:
            self.session.updateUserSettings(resp.raw['d'])
        #user session replaced (useful for syncing activities btwn client and server)
        elif resp.event.session_replaced:
            newStatus = resp.parsed.sessions_replace(session_id=self.session_id) #contains both status and activities
            self.session.updateUserSettings(newStatus)
    ######################################################

    '''
    Guild/Server stuff
    '''
    #op14 related stuff
    def getMemberFetchingParams(self, targetRangeStarts): #more for just proof of concept. targetRangeStarts must not contain duplicates and must be a list of integers
        targetRangeStarts = {i:1 for i in targetRangeStarts} #remove duplicates but preserve order
        if targetRangeStarts.get(0)!=None and targetRangeStarts.get(100)!=None:
            keys = list(targetRangeStarts)
            if keys.index(100)<keys.index(0):
                targetRangeStarts.pop(0) #needs to be removed or else fetchMembers will enter an infinite loop because of how discord responds to member list requests
        startIndex = 1 #can't start at 0 because can't divide by 0. No need to specify a stop index since fetchMembers continues until end of multipliers
        method = [0] #because startIndex is 1
        for index,i in enumerate(targetRangeStarts):
            method.append(i/(index+1))
        return startIndex, method #return startIndex and multipliers

    def fetchMembers(self, guild_id, channel_id, method="overlap", keep=[], considerUpdates=True, startIndex=0, stopIndex=1000000000, reset=True, wait=None, priority=0):
        if guild_id in self.memberFetchingStatus and reset:
            del self.memberFetchingStatus[guild_id] #just resetting tracker on the specific guild_id
        self.command(
            {
                "function": imports.GuildCombo(self).fetchMembers,
                "priority": priority,
                "params": {
                    "guild_id": guild_id,
                    "channel_id": channel_id,
                    "method": method,
                    "keep": keep,
                    "considerUpdates": considerUpdates,
                    "startIndex": startIndex,
                    "stopIndex": stopIndex,
                    "reset": reset,
                    "wait": wait
                },
            }
        )


    def finishedMemberFetching(self, guild_id):
        return self.memberFetchingStatus.get(guild_id) == "done"

    def findVisibleChannels(self, guildID, types=['guild_text', 'dm', 'guild_voice', 'group_dm', 'guild_category', 'guild_news', 'guild_store', 'guild_news_thread', 'guild_public_thread', 'guild_private_thread', 'guild_stage_voice'], findFirst=False):
        if len(self.session.read()[0]) == 0: #if never connected to gateway
            return
        return imports.GuildCombo(self).findVisibleChannels(guildID, types, findFirst)

    #sends a series of opcode 14s to tell discord that you're looking at guild channels
    def subscribeToGuildEvents(self, onlyLarge=False, wait=None):
        imports.GuildCombo(self).subscribeToGuildEvents(onlyLarge, wait)

    #op8 related stuff
    def queryGuildMembers(self, guildIDs, query, saveAsQueryOverride=None, limit=10, presences=True, keep=[]):
        if isinstance(guildIDs, str):
            guildIDs = [guildIDs]
        imports.GuildCombo(self).searchGuildMembers(guildIDs, query, saveAsQueryOverride, limit, presences, None, keep)

    def checkGuildMembers(self, guildIDs, userIDs, presences=True, keep=[]):
        if isinstance(guildIDs, str):
            guildIDs = [guildIDs]
        imports.GuildCombo(self).searchGuildMembers(guildIDs, "", None, 10, presences, userIDs, keep)

    def finishedGuildSearch(self, guildIDs, query="", saveAsQueryOverride=None, userIDs=None, keep=False):
        if isinstance(guildIDs, str):
            guildIDs = [guildIDs]
        saveAsQuery = query.lower() if saveAsQueryOverride==None else saveAsQueryOverride.lower()
        command = {
            "function": imports.GuildCombo(self).handleGuildMemberSearches,
            "params": {
                "guildIDs": guildIDs,
                "saveAsQuery": saveAsQuery,
                "isQueryOverridden": saveAsQueryOverride != None,
                "userIDs": userIDs,
                "keep": keep
            },
        }
        if keep == False: #if keep param not provided, look if params are subset of command_list function params
            command["params"].pop("keep")
            for c in self._after_message_hooks:
                if isinstance(c, dict):
                    if c.get("function").__func__ == imports.GuildCombo(self).handleGuildMemberSearches.__func__:
                        d1 = command["params"]
                        d2 = c.get("params", {})
                        if all(key in d2 and d2[key] == d1[key] for key in d1): #https://stackoverflow.com/a/41579450/14776493
                            return False #not finished yet with guild search
            return True
        else:
            return command not in self._after_message_hooks

    '''
    User stuff
    '''
    def setStatus(self, status): #can only be run while connected to gateway
        imports.User(self.RESTurl,self.sessionobj,self.log).setStatusHelper(status)
        imports.UserCombo(self).setStatus(status)

    def setPlayingStatus(self, game): #can only be run while connected to gateway, will update metadata later
        if not self.session.userSettings['show_current_game']:
            imports.User(self.RESTurl,self.sessionobj,self.log).enableActivityDisplay(enable=True)
        imports.UserCombo(self).setPlayingStatus(game)

    def removePlayingStatus(self): #can only be run while connected to gateway
        imports.UserCombo(self).removePlayingStatus()

    def setStreamingStatus(self, stream, url): #can only be run while connected to gateway, will update metadata later
        if not self.session.userSettings['show_current_game']:
            imports.User(self.RESTurl,self.sessionobj,self.log).enableActivityDisplay(enable=True)
        imports.UserCombo(self).setStreamingStatus(stream, url)

    def removeStreamingStatus(self): #can only be run while connected to gateway
        imports.UserCombo(self).removeStreamingStatus()

    def setListeningStatus(self, song): #can only be run while connected to gateway, will update metadata later
        if not self.session.userSettings['show_current_game']:
            imports.User(self.RESTurl,self.sessionobj,self.log).enableActivityDisplay(enable=True)
        imports.UserCombo(self).setListeningStatus(song)

    def removeListeningStatus(self): #can only be run while connected to gateway
        imports.UserCombo(self).removeListeningStatus()

    def setWatchingStatus(self, show): #can only be run while connected to gateway, will update metadata later
        if not self.session.userSettings['show_current_game']:
            imports.User(self.RESTurl,self.sessionobj,self.log).enableActivityDisplay(enable=True)
        imports.UserCombo(self).setWatchingStatus(show)

    def removeWatchingStatus(self): #can only be run while connected to gateway
        imports.UserCombo(self).removeWatchingStatus()

    def setCustomStatus(self, customstatus, emoji=None, animatedEmoji=False, expires_at=None): #can only be run while connected to gateway
        imports.User(self.RESTurl,self.sessionobj,self.log).setStatusHelper(self.session.userSettings['status'])
        imports.User(self.RESTurl,self.sessionobj,self.log).setCustomStatusHelper(customstatus, emoji, expires_at)
        imports.UserCombo(self).setCustomStatus(customstatus, emoji, animatedEmoji)

    def removeCustomStatus(self):
        imports.User(self.RESTurl,self.sessionobj,self.log).setCustomStatusHelper("")
        imports.UserCombo(self).removeCustomStatus()

    def clearActivities(self):
        if self.session.userSettings['custom_status'] != None:
            imports.User(self.RESTurl,self.sessionobj,self.log).setCustomStatusHelper("", emoji=None, expires_at=None)
        imports.UserCombo(self).clearActivities()



#points to commands that help request info/actions using the gateway

class GuildRequest(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def lazyGuild(self, guild_id, channel_ranges, typing, threads, activities, members, thread_member_lists): #https://arandomnewaccount.gitlab.io/discord-unofficial-docs/lazy_guilds.html
		data = {
		    "op": self.gatewayobj.OPCODE.LAZY_REQUEST,
		    "d": {
		        "guild_id": guild_id,
		        "typing": typing,
		        "threads": threads,
		        "activities": activities,
		        "members": members,
		        "channels": channel_ranges,
		        "thread_member_lists": thread_member_lists
		    },
		}
		if channel_ranges == None:
			data["d"].pop("channels")
		if typing == None:
			data["d"].pop("typing")
		if threads == None:
			data["d"].pop("threads")
		if activities == None:
			data["d"].pop("activities")
		if members == None:
			data["d"].pop("members")
		if thread_member_lists == None:
			data["d"].pop("thread_member_lists")
		self.gatewayobj.send(data)

	def searchGuildMembers(self, guild_ids, query, limit, presences, user_ids, nonce): #note that query can only be "" if you have admin perms (otherwise you'll get inconsistent responses from discord)
		if isinstance(guild_ids, str):
			guild_ids = [guild_ids]
		data = {
		    "op": self.gatewayobj.OPCODE.REQUEST_GUILD_MEMBERS,
		    "d": {"guild_id": guild_ids},
		}
		if isinstance(user_ids, list): #there are 2 types of op8 that the client can send
			data["d"]["user_ids"] = user_ids
		else:
			data["d"]["query"] = query
			data["d"]["limit"] = limit
		if presences != None:
			data["d"]["presences"] = presences
		if nonce != None:
			data["d"]["nonce"] = nonce
		self.gatewayobj.send(data)

	#wait but why is this under guild? because gateway slash command searching only works in guilds
	#example "d" fields (from client):
	'''
	{"guild_id": "blah", "nonce": "blah", "type": 1, "limit": 7, "query": "test"}
	{"guild_id": "blah", "nonce": "blah", "type": 1, "applications": True, "offset": 0, "command_ids": ["blah", ...], "limit":10}
	{"guild_id": "blah", "nonce": "blah", "type": 1, "offset": 30, "limit": 10}
	'''
	def searchSlashCommands(self, guild_id, query, command_ids, applicationID, limit, offset, nonce, appType):
		#nonce
		if nonce == "calculate":
			from ...utils.nonce import calculateNonce
			nonce = calculateNonce()
		else:
			nonce = str(nonce)

		#payload
		data = {
			"op": self.gatewayobj.OPCODE.REQUEST_APPLICATION_COMMANDS,
			"d": {"guild_id": guild_id, "nonce": nonce, "type": 1},
		}

		#application type
		appType = appType.lower()
		if 'user' in appType:
			data["d"]["type"] = 2
		elif 'message' in appType:
			data["d"]["type"] = 3

		#case 1: search by query
		if query != None:
			data["d"].update({"limit":limit, "query":query})
			if offset != None:
				data["d"]["offset"] = offset
		#case 2: search by command ids
		elif command_ids != None:
			if isinstance(command_ids, str):
				command_ids = [command_ids]
			data["d"].update({"applications":True, "offset":0 if offset==None else offset, "command_ids":command_ids, "limit":limit})
		#case 3: get them all (with a limit)
		else:
			data["d"]["limit"] = limit
			if offset != None:
				data["d"]["offset"] = offset

		#idk, this isn't ever done in the client but it's useful so here goes
		if applicationID != None:
			data["d"]["application_id"] = applicationID
		self.gatewayobj.send(data)


from ..types import Types

#parse
class MessageParse(object):
	@staticmethod
	def message_create(response):
		message = dict(response["d"])
		message["type"] = Types.MESSAGE[response["d"]["type"]] #number to str
		if "member" in message:
			extra_data = message.pop("member")
			message["author"].update(extra_data)
		return message


'''
sources used:
https://luna.gitlab.io/discord-unofficial-docs/desktop_remote_auth.html
https://github.com/Vap0r1ze/discord-remote-auth/blob/master/src/types/RemoteAuthClient.js
'''
import websocket
import time

import json
import base64

try:
	from Crypto.PublicKey import RSA
	from Crypto.Cipher import PKCS1_OAEP
	from Crypto.Hash import SHA256
	import pyqrcode
except ImportError:
	raise ImportError("Install discum[ra] to use remote authentication.")

try:
	import thread
except ImportError:
	import _thread as thread

from ..logger import *

class RemoteAuth:

	__slots__ = ['user_agent', 'saveQrCode', 'proxy_host', 'proxy_port', 'log', '_last_err', '_after_message_hooks', 'connected', 'interval', 'key_pair', 'public_key', 'decryptor', 'fingerprint', 'qr_url', 'qr_img', 'userData', 'token', 'ws']

	class OPCODE:
		# Name                 Code                      Client Action   Description
		HELLO =                "hello"  #                Receive         sent on connection open
		INIT =                 "init"  #                 Send            describes generated public key
		NONCE_PROOF =          "nonce_proof"  #          Send/Receive    encrypted nonce from server, decrypted nonce from client as "proof"
		PENDING_REMOTE_INIT =  "pending_remote_init"  #  Receive         sent after a valid nonce_proof is submitted
		PENDING_FINISH =       "pending_finish"  #       Receive         sent after QR code is scanned, contains encrypted user data
		FINISH =               "finish"  #               Receive         sent after login flow is completed, contains encrypted token
		CANCEL =               "cancel"  #               Receive         cancel qr code login
		HEARTBEAT =            "heartbeat"  #            Send            sent every N ms, described in hello packet
		HEARTBEAT_ACK =        "heartbeat_ack"  #        Receive         sent after heartbeat packet, should close connection if a heartbeat_ack isn't received by the next heartbeat interval

	def __init__(self, remoteauthurl, user_agent, proxy_host=None, proxy_port=None, log={"console":True, "file":False}):
		self.user_agent = user_agent
		self.saveQrCode = True
		self.proxy_host = proxy_host
		self.proxy_port = proxy_port
		self.log = log
		self._last_err = None
		self._after_message_hooks = []
		self.connected = False
		self.interval = None
		self.key_pair = None
		self.public_key = None
		self.decryptor = None
		self.fingerprint = None
		self.qr_url = None
		self.qr_img = None
		self.userData = None
		self.token = None
		self.ws = self._get_ws_app(remoteauthurl)

	#WebSocketApp, more info here: https://github.com/websocket-client/websocket-client/blob/master/websocket/_app.py#L84
	def _get_ws_app(self, remoteauthurl):
		headers = {
			"Accept-Encoding": "gzip, deflate, br",
			"Accept-Language": "en-US,en;q=0.9",
			"Cache-Control": "no-cache",
			"Pragma": "no-cache",
			"Sec-WebSocket-Extensions": "permessage-deflate; client_max_window_bits",
			"User-Agent": self.user_agent
		} #more info: https://stackoverflow.com/a/40675547

		ws = websocket.WebSocketApp(remoteauthurl,
									header = headers,
									on_open=lambda ws: self.on_open(ws),
									on_message=lambda ws, msg: self.on_message(ws, msg),
									on_error=lambda ws, msg: self.on_error(ws, msg),
									on_close=lambda ws, close_code, close_msg: self.on_close(ws, close_code, close_msg)
									)
		return ws

	def decrypt(self, data):
		decoded = base64.b64decode(data)
		decrypted = self.decryptor.decrypt(decoded)
		return decrypted

	def parseUserPayload(self, data):
		user_data = data.decode('utf-8').split(':')
		userID = user_data[0]
		username = user_data[3]
		discriminator = user_data[1]
		avatarName = user_data[2]
		if avatarName == '0':
			avatar = 'https://cdn.discordapp.com/embed/avatars/{}.png'.format(int(discriminator) % 5)
		elif avatarName.startwith('a_'):
			avatar = 'https://cdn.discordapp.com/avatars/{}/{}.gif'.format(userID, avatarName)
		else:
			avatar = 'https://cdn.discordapp.com/avatars/{}/{}.png'.format(userID, avatarName)
		return userID, username, discriminator, avatar

	def on_open(self, ws):
		self.connected = True
		Logger.log("[ra] Connected to websocket.", None, self.log)

	def on_message(self, ws, message):
		response = json.loads(message)
		Logger.log('[ra] < {}'.format(response), LogLevel.RECEIVE, self.log)
		op = response['op']
		if op == self.OPCODE.HELLO:
			self.interval = (response["heartbeat_interval"])/1000
			thread.start_new_thread(self._heartbeat, ())
			cert = self.public_key.exportKey()
			encoded_public_key = cert.decode('utf-8').replace('\n','')[26:-24]
			self.send({"op":self.OPCODE.INIT, "encoded_public_key":encoded_public_key})
		elif op == self.OPCODE.NONCE_PROOF:
			nonce = self.decrypt(response['encrypted_nonce'])
			digest = SHA256.new(nonce).digest()
			nonce_proof = base64.urlsafe_b64encode(digest).decode('utf-8').rstrip('=')
			self.send({"op": self.OPCODE.NONCE_PROOF, "proof": nonce_proof})
		elif op == self.OPCODE.PENDING_REMOTE_INIT:
			self.fingerprint = response['fingerprint']
			self.qr_url = 'https://discordapp.com/ra/{}'.format(self.fingerprint)
			self.qr_img = pyqrcode.create(self.qr_url, error='H')
			if self.saveQrCode: #self.qr_img.show() wasn't working consistently
				if isinstance(self.saveQrCode, str):
					fileLoc = self.saveQrCode
				else:
					fileLoc = self.fingerprint+".png"
				self.qr_img.png(fileLoc, scale=10)
				print("QR code image for {} saved in {}".format(self.fingerprint, fileLoc))
		elif op == self.OPCODE.PENDING_FINISH:
			user_payload = self.decrypt(response['encrypted_user_payload'])
			userID, username, discriminator, avatar = self.parseUserPayload(user_payload)
			self.userData = {
				'id': userID,
				'username': username,
				'discriminator': discriminator,
				'avatar': avatar
			}
		elif response['op'] == self.OPCODE.FINISH:
			self.token = self.decrypt(response['encrypted_token']).decode('utf-8')
		if self.interval == None:
			Logger.log("[ra] Connection failed.", None, self.log)
			self.close()
		thread.start_new_thread(self._response_loop, (response,))

	def on_error(self, ws, error):
		Logger.log('[ra] < {}'.format(error), LogLevel.WARNING, self.log)
		self._last_err = error

	def on_close(self, ws, close_code, close_msg):
		self.connected = False
		if close_code or close_msg:
			self._last_close_event = {"code": close_code, "reason": close_msg}
			print(self._last_close_event)
		Logger.log('[ra] websocket closed', None, self.log)

	def _heartbeat(self):
		Logger.log("[ra] entering heartbeat", None, self.log)
		while self.connected:
			time.sleep(self.interval)
			if not self.connected:
				break
			self.send({"op": self.OPCODE.HEARTBEAT})
		return

	#just a wrapper for ws.send
	def send(self, payload):
		Logger.log('[ra] > {}'.format(payload), LogLevel.SEND, self.log)
		self.ws.send(json.dumps(payload))

	def close(self):
		self.connected = False
		Logger.log('[ra] websocket closed', None, self.log)
		self.ws.close()

	def command(self, func):
		if callable(func):
			self._after_message_hooks.append(func)
			return func
		elif isinstance(func, dict):
			priority = func.pop('priority', len(self._after_message_hooks))
			self._after_message_hooks.insert(priority, func)
			return func['function']

	#kinda influenced by https://github.com/scrubjay55/Reddit_ChatBot_Python (Apache License 2.0)
	def _response_loop(self, response):
		commandslist = self._after_message_hooks[:] #create a copy
		for func in commandslist:
			if callable(func):
				func(response)
			elif isinstance(func, dict):
				function = func['function']
				params = func['params'] if 'params' in func else {}
				function(response, **params)
		return

	def removeCommand(self, func, exactMatch=True, allMatches=False):
		try:
			if exactMatch:
				self._after_message_hooks.index(func) #for raising the value error
				if allMatches:
					self._after_message_hooks = [i for i in self._after_message_hooks if i!=func]
				else: #simply remove first found
					del self._after_message_hooks[self._after_message_hooks.index(func)]
			else:
				commandsCopy = [i if callable(i) else i['function'] for i in self._after_message_hooks] #list of just functions
				commandsCopy.index(func) #for raising the value error
				if allMatches:
					self._after_message_hooks = [i for (i,j) in zip(self._after_message_hooks, commandsCopy) if j!=func]
				else:
					del self._after_message_hooks[commandsCopy.index(func)]
		except ValueError:
			Logger.log('{} not found in _after_message_hooks.'.format(func), None, self.log)
			pass

	def clearCommands(self):
		self._after_message_hooks = []

	def resetSession(self):
		self._last_err = None
		self.connected = False
		self.interval = None
		self.key_pair = RSA.generate(2048)
		self.public_key = self.key_pair.publickey()
		self.decryptor = PKCS1_OAEP.new(self.key_pair, SHA256)
		self.fingerprint = None
		self.qr_url = None
		self.qr_img = None
		self.userData = None
		self.token = None

	def run(self, saveQrCode=True):
		self.saveQrCode = saveQrCode
		self.resetSession()
		self.ws.run_forever(ping_interval=10, ping_timeout=5, http_proxy_host=self.proxy_host, http_proxy_port=self.proxy_port)
		return self.token, self.userData